/****************************************************************************\
*            
*     FILE:     DataGridView.c
*
*     PURPOSE:  ListView based Datagrid with Editable subItems
*
*     COMMENTS:
*               This source is distributed in the hope that it will be useful,
*               but WITHOUT ANY WARRANTY; without even the implied warranty of
*               MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*
*     FUNCTIONS:
*      EXPORTS:
*               InitDataGridView() - Initialize and register the custom control 
*
*               ToDo: add export functions here
*      LOCALS:
*               Grid_GetInstanceData() - Get the Instance data associated with this instance.
*               Grid_CreateInstanceData() - Allocate the Instance data associated with this instance.
*               Grid_FreeInstanceData() - Free the instance data allocation of an instance of the Grid Control.
*               Grid_GetColumnCount() - Under any condition get the number of columns in grid.
*               ListView_GetFirstVisibleRow() - Get the row number of the first visible row.
*               ListView_GetLastVisibleRow() - Get the row number of the last visible row.
*               ListView_GetRowHeight() - Get the hight of the list view component rows.
*               ListView_GetLastVisibleColWidth() - Get the width of the last visible col.
*               ListView_GetFirstVisibleColWidth() - Get the width of the first visible col.
*               SetColor() - Set the colors used to paint controls in OnCtlColor....
*               Grid_OnCtlColorEdit() - Set the colors of the Editor's Edit Control
*				ListView_Resize() - Resize List View within control
*				ResizeEditor() - Resize Editor within control (initially)
*				CreateListView() - Creates the List View
*				CreateCellEditor() - Creates the Cell Editor
*               Grid_SetRowHeight() - Hack - Use image list to set row height
*               Grid_OnCreate() - Handles WM_CREATE
*               Grid_OnDestroy() - Handles WM_DESTROY
*               Grid_OnGetDlgCode() - Handles WM_GETDLGCODE sent by a dialog procedure if control is
*                                      used in a dialog.
*               Grid_OnSize() - Handles WM_SIZE
*               SetWindowScrolls() - Initializes the scroll bars and the scroll info for the window
*               Scroll() - Generic wrapper for ScrollWindowEx
*               Grid_OnHScroll() - Handle the horizontal scrollbar
*               Grid_OnVScroll() - Handle the vertical scrollbar
*               Grid_OnSetCursor() - The cursor changes when the mouse leaves the header, this is the
*                                     only easy indication that a column resize has been compleated.
*                                    I use the event to crop the grid back to its data and to reposition
*                                     the editor
*               Grid_OnColumnResize() - Handles the HDN_ITEMCHANGING notification due to a resize attempt
*               Grid_OnCustomDraw() - Define the look and feel of the Grid
*               Grid_OnMouseClick() - Handles mouse clicks on the ListView
*               Grid_OnNotify() - Handles notification messages for control's components
*               Grid_OnCommand() - Handles notification codes of control's components
*               ListView_KeyProc() - Window Procedure for the ListView
*               Editor_KeyProc() - Window Procedure for the Editor
*               CbEditor_ComboProc() - Window Procedure for the ComboBoxEditor
*               ListViewHeader_MoveProc() - Window Procedure for the ListView Header
*               Grid_WndProc() - Window Procedure for the Data View Grid
*
*     Copyright 2007 David MacDermot.
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "DataGridView.h"

#define ID_LISTVIEW  2000
#define ID_EDIT 2001

/****************************************************************************/
// DataGridView.c Macroe Definitions

#define NELEMS(a) (sizeof(a) / sizeof((a)[0]))

#define Refresh(hWnd) RedrawWindow(hWnd,NULL,NULL,RDW_NOERASE|RDW_INVALIDATE|RDW_ALLCHILDREN|RDW_UPDATENOW);

#define PinHeaderToTop(hHeader)	SNDMSG((hHeader), WM_MOVE, (WPARAM)0, (LPARAM)0L)

/****************************************************************************/
// DataGridView.c Structure

typedef struct _tagINSTANCEDATA{
	HINSTANCE hInstance;
	HWND hwndList;
	HWND hwndEditor;
	LVHITTESTINFO hti;
	SCROLLINFO hsInfo;
	SCROLLINFO vsInfo;
	COLORREF Editor_TxtColr;
	COLORREF Editor_BkColr;
	COLORREF Alt_TxtColr;
	COLORREF Alt_BkColr;
	BOOL fPaintByRow;
	BOOL fRowHeaders;
	BOOL fResizableHeader;
	BOOL fDblClick;
}INSTANCEDATA, *LPINSTANCEDATA;

/****************************************************************************/
// DataGridView.c Internal Globals

LPINSTANCEDATA g_lpInst;
TCHAR g_szClassName[] = TEXT("DGridVwClass");
BOOL g_fsizeCol;

/****************************************************************************/
// DataGridView.c Function prototypes

static LRESULT CALLBACK Grid_WndProc (HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK Editor_KeyProc (HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK ListView_KeyProc (HWND, UINT, WPARAM, LPARAM);
static LRESULT CALLBACK ListViewHeader_MoveProc (HWND, UINT, WPARAM, LPARAM);
static VOID SetWindowScrolls(HWND, PRECT);

/****************************************************************************
*
*     FUNCTION: Grid_GetInstanceData
*
*     PURPOSE:  Get the Instance data associated with this instance.
*
*     PARAMS:   HWND hGrid - Handle to Current instance
*               LPINSTANCEDATA *ppInstanceData - pointer to the address of an INSTANCEDATA struct
*
*     RETURNS:  BOOL - TRUE if successful
*
* History:
*                August '07 - Created
*
\****************************************************************************/

static BOOL Grid_GetInstanceData (HWND hGrid, LPINSTANCEDATA *ppInstanceData)
{
	*ppInstanceData=(LPINSTANCEDATA)GetProp(hGrid,(LPCSTR)"lpInsData");
	if(NULL != *ppInstanceData) return TRUE;
	return FALSE; 
}

/****************************************************************************
*
*     FUNCTION: Grid_CreateInstanceData
*
*     PURPOSE:  Allocate the Instance data associated with this instance.
*
*     PARAMS:   HWND hGrid - Handle to Current instance
*               LPINSTANCEDATA pInstanceData - pointer to an INSTANCEDATA struct
*
*     RETURNS:  BOOL - TRUE if successful
*
* History:
*                August '07 - Created
*
\****************************************************************************/

static BOOL Grid_CreateInstanceData (HWND hGrid,  LPINSTANCEDATA pInstanceData)
{
	LPINSTANCEDATA pInst = (LPINSTANCEDATA)malloc(sizeof(INSTANCEDATA));
	memmove(pInst,pInstanceData,sizeof(INSTANCEDATA));

	return SetProp(hGrid,"lpInsData",pInst);
}

/****************************************************************************
*
*     FUNCTION: Grid_FreeInstanceData
*
*     PURPOSE:  Free the instance data allocation of an instance of the Grid Control.
*
*     PARAMS:   HWND hGrid - Handle to Current instance
*
*     RETURNS:  BOOL - TRUE if successful
*
* History:
*                August '07 - Created
*
\****************************************************************************/

static BOOL Grid_FreeInstanceData (HWND hGrid)
{
	LPINSTANCEDATA pInst;
	if(Grid_GetInstanceData(hGrid,&pInst))
	{
		free ((LPINSTANCEDATA)pInst);
		RemoveProp(hGrid,"lpInsData");
		return TRUE;
	}
	return FALSE;
}

/****************************************************************************
*
*     FUNCTION: Grid_GetColumnCount
*
*     PURPOSE:  Under any condition get the number of columns in grid.
*
*     PARAMS:   HWND hwnd - Handle of Grid
*
*     RETURNS:  INT - number of columns
*
* History:
*                August '07 - Created
*
\****************************************************************************/

static INT Grid_GetColumnCount(HWND hwnd)
{
	// Method: hittest low right corner of row
	BOOL fTempRow = FALSE;
	RECT rc;
	LVHITTESTINFO ht = {0};

	if(0 == ListView_GetItemCount(hwnd)) // No rows yet
	// Temporarily add a row in order to get the count
	{
		LV_ITEM Lv_i = {LVIF_TEXT,0,0,0,0,""};
		ListView_InsertItem(hwnd,&Lv_i);
		fTempRow = TRUE;
	}
	ListView_GetItemRect(hwnd,
		ListView_GetItemCount(hwnd)-1,
		&rc,LVIR_BOUNDS);
	ht.pt.x = rc.right-1;
	ht.pt.y = rc.bottom-1;
	ListView_SubItemHitTest(hwnd, &ht);

	if(fTempRow) ListView_DeleteItem(hwnd,0); //Delete temp row (if any)
		
	return ht.iSubItem;
}

/****************************************************************************
*
*     FUNCTION: ListView_GetLastVisibleRow
*
*     PURPOSE:  Get the row number of the last visible row.
*
*     PARAMS:   HWND hwnd - Handle of ListView
*
*               PRECT lprc - pointer to rectangle (Grid Control)
*
*     RETURNS:  INT - number of last row
*
* History:
*                Sept '07 - Created
*
\****************************************************************************/

static INT ListView_GetLastVisibleRow(HWND hList, PRECT lprc)
{
	LVHITTESTINFO hti = {0};
	hti.pt.x = lprc->left + 10;
	hti.pt.y = lprc->bottom;
	MapWindowPoints(GetParent(hList), hList, &hti.pt,1);

	ListView_SubItemHitTest(hList, &hti);

	return hti.iItem;
}

/****************************************************************************
*
*     FUNCTION: ListView_GetRowHeight
*
*     PURPOSE:  Get the hight of the list view component rows.
*
*     PARAMS:   HWND hwnd - Handle of ListView
*
*               PRECT lprc - pointer to rectangle (Grid Control)
*
*     RETURNS:  INT - height of first row
*
* History:
*                Sept '07 - Created
*
\****************************************************************************/

static INT ListView_GetRowHeight(HWND hList, PRECT lprc)
{
	RECT rc;
	if(ListView_GetItemRect(hList,0,&rc,LVIR_BOUNDS))
		return rc.bottom-rc.top;
	else
		return 0;
}

/****************************************************************************
*
*     FUNCTION: ListView_GetLastVisibleColWidth
*
*     PURPOSE:  Get the width of the last visible col.
*
*     PARAMS:   HWND hwnd - Handle of ListView
*
*               PRECT lprc - pointer to rectangle (Grid Control)
*
*     RETURNS:  INT - width of last col
*
* History:
*                Sept '07 - Created
*
\****************************************************************************/

static INT ListView_GetLastVisibleColWidth(HWND hList, PRECT lprc)
{
	LVHITTESTINFO hti = {0};
	hti.pt.x = lprc->right;
	hti.pt.y = lprc->bottom;
	MapWindowPoints(GetParent(hList), hList, &hti.pt,1);
	ListView_SubItemHitTest(hList, &hti);

	if(hti.iSubItem < 0)
		return 0;
	else
		return ListView_GetColumnWidth(hList,hti.iSubItem);
}

/****************************************************************************
*
*     FUNCTION: ListView_GetFirstVisibleColWidth
*
*     PURPOSE:  Get the width of the first visible col.
*
*     PARAMS:   HWND hwnd - Handle of ListView
*
*               PRECT lprc - pointer to rectangle (Grid Control)
*
*     RETURNS:  INT - width of first col
*
* History:
*                Sept '07 - Created
*
\****************************************************************************/

static INT ListView_GetFirstVisibleColWidth(HWND hList, PRECT lprc)
{
	LVHITTESTINFO hti = {0};
	hti.pt.x = lprc->left + 10;
	hti.pt.y = lprc->top;
	MapWindowPoints(GetParent(hList), hList, &hti.pt,1);
	ListView_SubItemHitTest(hList, &hti);

	if(hti.iSubItem < 0)
		return 0;
	else
		return ListView_GetColumnWidth(hList,hti.iSubItem);
}

/****************************************************************************
*
*     FUNCTION: SetColor
*
*     PURPOSE:  Set the colors used to paint controls in OnCtlColor....
*
*     PARAMS:   COLORREF TxtColr - Desired text color
*               COLORREF BkColr - Desired back color
*               HDC hdc - Handle of a device context
*
*     RETURNS:  HBRUSH - A reusable brush object
*
* History:
*                August '07 - Created
*
\****************************************************************************/

HBRUSH SetColor (COLORREF TxtColr, COLORREF BkColr, HDC hdc)
{
	static HBRUSH  ReUsableBrush;
	DeleteObject(ReUsableBrush);
	ReUsableBrush=CreateSolidBrush(BkColr);
	SetTextColor(hdc,TxtColr);
	SetBkColor(hdc,BkColr);
	return ReUsableBrush;
}

/****************************************************************************
*
*     FUNCTION: Grid_OnCtlColorEdit
*
*     PURPOSE:  Set the colors of the Editor's Edit Control
*
*     PARAMS:   HWND hwnd - Handle of Grid
*               HDC hdc - Handle of a device context
*               HWND hwndChild - Handle of the Edit Control to custom paint
*               INT type - CTLCOLOR_EDIT
*
*     RETURNS:  HBRUSH - A reusable brush object
*
* History:
*                August '07 - Created
*
\****************************************************************************/

HBRUSH Grid_OnCtlColorEdit(HWND hwnd, HDC hdc, HWND hwndChild, INT type)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);

	if(g_lpInst->hwndEditor == hwndChild)
		return SetColor(g_lpInst->Editor_TxtColr,g_lpInst->Editor_BkColr,hdc);
	return NULL;
}

/****************************************************************************
*
*     FUNCTION: CreateListView
*
*     PARAMS:   HINSTANCE hInstance - handle of application instance
*               HWND hwndParent - handle of list view parent
*
*     RETURNS:  HWND - the handle of the list view
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static HWND CreateListView(HINSTANCE hInstance, HWND hwndParent)
{
    DWORD dwStyle, dwExStyle;
    HWND hwnd;

	// Create the ListView
	dwStyle = WS_CHILD|
		WS_VISIBLE|
		LVS_REPORT|
		LVS_SINGLESEL;

	dwExStyle = WS_EX_LEFT;

    hwnd = CreateWindowEx(dwExStyle,       // ex style
        WC_LISTVIEW,                       // class name - defined in commctrl.h
        TEXT(""),                          // dummy text
        dwStyle,                           // style
        0,                                 // x position
        0,                                 // y position
        0,                                 // width
        0,                                 // height
        hwndParent,                        // parent
        (HMENU)ID_LISTVIEW,                // ID
        hInstance,                         // instance
        NULL);                             // no extra data

    if (!hwnd) return NULL;

	ListView_SetExtendedListViewStyle(hwnd,LVS_EX_GRIDLINES);
	SendMessage(hwnd, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT),0);

	// Subclass ListView and save the OldProc
	SetProp(hwnd,"Wprc",(HANDLE)GetWindowLong(hwnd,GWL_WNDPROC));
	SubclassWindow(hwnd,ListView_KeyProc);

	// Subclass ListView Header
	HWND hHeader = ListView_GetHeader(hwnd);

	SetProp(hHeader,"Wprc",(HANDLE)GetWindowLong(hHeader,GWL_WNDPROC));
	SubclassWindow(hHeader,ListViewHeader_MoveProc);

    return hwnd;
}

/****************************************************************************
*
*     FUNCTION: CreateCellEditor
*
*     PARAMS:   HINSTANCE hInstance - handle of application instance
*               HWND hwndParent - handle of list view parent
*
*     RETURNS:  HWND - the handle of the CellEditor
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static HWND CreateCellEditor(HINSTANCE hInstance, HWND hwndParent)
{
    DWORD dwStyle, dwExStyle;
    HWND hwnd;

    dwStyle = WS_CHILD|
		WS_BORDER|
		ES_LEFT|
		ES_MULTILINE|
		ES_WANTRETURN;

	dwExStyle = WS_EX_TRANSPARENT|
		WS_EX_LEFT;

    hwnd = CreateWindowEx(dwExStyle,// ex style
        WC_EDIT,                    // class name - defined in commctrl.h
        TEXT(""),                   // dummy text
        dwStyle,                    // style
        0,                          // x position
        0,                          // y position
        0,                          // width
        0,                          // height
        hwndParent,                 // parent
        (HMENU)ID_EDIT,             // ID
        hInstance,                  // instance
        NULL);                      // no extra data

    if (!hwnd) return NULL;

	SendMessage(hwnd, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT),0);

	// Subclass Editor and save the OldProc
	SetProp(hwnd,"Wprc",(HANDLE)GetWindowLong(hwnd,GWL_WNDPROC));
	SubclassWindow(hwnd,Editor_KeyProc);

    return hwnd;
}

/****************************************************************************
*
*     FUNCTION: Grid_SetRowHeight
*
*     PURPOSE:  Hack - Use image list to set row height.
*
*     PARAMS:   HWND hList - Handle to listview
*               INT iRowHeight - Desired row height
*
*     RETURNS:  VOID
*
* History:
*                Sep '07 - Created
*
\****************************************************************************/

VOID Grid_SetRowHeight(HWND hList, INT iRowHeight)
{
	static HIMAGELIST hIconList;
	ImageList_Destroy (hIconList);
	hIconList = ImageList_Create(1,iRowHeight,ILC_COLOR,0,1);
	ListView_SetImageList(hList,hIconList,LVSIL_SMALL);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnCreate
*
*     PURPOSE:  Handles WM_CREATE
*
*     PARAMS:   HWND hwnd - handle of Control
*               LPCREATESTRUCT lpCreateStruct - structure with creation data
*
*     RETURNS:  BOOL - If an application processes this message,
*                      it should return TRUE to continue creation of the window.
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static BOOL Grid_OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	INSTANCEDATA inst;

	// initialize hit test info and scroll info
	memset(&inst.hti, -1, sizeof(LVHITTESTINFO));
	memset(&inst.hsInfo, 0, sizeof(SCROLLINFO));
	memset(&inst.vsInfo, 0, sizeof(SCROLLINFO));

	// get the hInstance
	inst.hInstance = lpCreateStruct->hInstance;

    // create the ListView control
    inst.hwndList = CreateListView(lpCreateStruct->hInstance, hwnd);
	if(NULL == inst.hwndList) return FALSE;

	// default ListView Colors
	inst.fPaintByRow = TRUE;
	inst.Alt_TxtColr = ListView_GetTextColor(inst.hwndList);
	inst.Alt_BkColr = ListView_GetBkColor(inst.hwndList);

	// default ListView pseudoHeaders off
	inst.fRowHeaders = FALSE;

    inst.hwndEditor = CreateCellEditor(lpCreateStruct->hInstance, hwnd);
	if(NULL == inst.hwndEditor) return FALSE;

	// default Cell Editor Colors
	inst.Editor_BkColr = RGB(0,0,160);
	inst.Editor_TxtColr = RGB(255,255,255);

	// default Cell selection behavior
	inst.fDblClick = FALSE;

	Grid_SetRowHeight(inst.hwndList, 20);

	// TODO: Create inializations for an array of cell edit controls

	return Grid_CreateInstanceData(hwnd,&inst);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnDestroy
*
*     PURPOSE:  Handles WM_DESTROY
*
*     PARAMS:   HWND hwnd - handle of Control
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID Grid_OnDestroy(HWND hwnd)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);

	DestroyWindow(g_lpInst->hwndList);
	DestroyWindow(g_lpInst->hwndEditor);

	Grid_FreeInstanceData(hwnd);
	PostQuitMessage(0);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnGetDlgCode
*
*     PURPOSE:  Handles WM_GETDLGCODE sent by a dialog procedure if control is
*               used in a dialog.
*
*     PARAMS:   HWND hwnd - handle of control
*               LPMSG lpmsg - If lpmsg is not NULL, it is a far pointer to an MSG
*                			  structure that contains a message that is being sent to
*                			  the control.  Keyboard messages Include WM_KEYDOWN,
*							  WM_SYSCHAR, and WM_CHAR.
*
*     RETURNS:  BOOL - TRUE if handled
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static UINT Grid_OnGetDlgCode(HWND hwnd, LPMSG lpmsg)
{
	return DLGC_WANTALLKEYS;
}

/****************************************************************************
*
*     FUNCTION: ListView_Resize
*
*     PURPOSE:  Special resize for use when user resizes columns manually
*               prevents the scroll bars from showing.
*     PARAMS:   HWND hwndListView - handle of the list view
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID ListView_Resize(HWND hwndListView)
{
	RECT rc;
	ListView_GetItemRect(hwndListView, ListView_GetItemCount(hwndListView)-1,
			&rc,LVIR_BOUNDS);

	SetWindowScrolls(GetParent(hwndListView), &rc);

	//Size ListView component slightly larger than its contents
	SetWindowPos(hwndListView,NULL, 0, 0,
		rc.right - rc.left+20, rc.bottom+12,SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnSize
*
*     PURPOSE:  Handles WM_SIZE
*
*     PARAMS:   HWND hwnd - handle of Control
*               UINT state - Specifies the type of resizing requested
*               INT cx - width of client area
*               INT cy - height of client area
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID Grid_OnSize(HWND hwnd, UINT state, INT cx, INT cy)
{
	RECT rc;
	Grid_GetInstanceData(hwnd,&g_lpInst);

	ListView_GetItemRect(g_lpInst->hwndList,
			ListView_GetItemCount(g_lpInst->hwndList)-1,
			&rc,LVIR_BOUNDS);

	SetWindowScrolls(hwnd, &rc);

	//Size ListView component to its contents
	 SetWindowPos(g_lpInst->hwndList,NULL, 0, 0,
		 rc.right - rc.left, rc.bottom,SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
}

/****************************************************************************
*
*     FUNCTION: Scroll
*
*     PURPOSE:  Generic wrapper for ScrollWindowEx.
*
*     COMMENT:  Called by both Grid_OnVScroll and Grid_OnHScroll.
*
*     PARAMS:   HWND hwnd	- Handle of window to scroll
*               INT dx		- Desired new x position
*               INT dy		- Desired new y position
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID Scroll(HWND hwnd,INT dx,INT dy)
{
	ScrollWindowEx(
		hwnd, // handle of window to scroll
		dx,   // amount of horizontal scrolling
		dy,   // amount of vertical scrolling
		NULL, // address of structure with scroll rectangle
		NULL, // address of structure with clip rectangle
		NULL, // handle of update region
		NULL, // address of structure for update rectangle
		SW_SCROLLCHILDREN | SW_INVALIDATE // scrolling flags
		);
}

/****************************************************************************
*
*     FUNCTION: SetWindowScrolls
*
*     PURPOSE:  Set / update the scroll bars and the scroll info for the window.
*
*     COMMENT:  Should be called when handling WM_SIZE.
*
*      PARAMS:  HWND hwnd	- Handle of window to scroll
*               PRECT prc	- Pointer to window size rectangle
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID SetWindowScrolls(HWND hwnd, PRECT prc)
{
	RECT rc;
	GetClientRect(hwnd,&rc);
	Grid_GetInstanceData(hwnd,&g_lpInst);

	g_lpInst->hsInfo.fMask = SIF_ALL;
	g_lpInst->hsInfo.cbSize = sizeof(SCROLLINFO);
	g_lpInst->hsInfo.nMin = 0;
	g_lpInst->hsInfo.nMax = prc->right;
	g_lpInst->hsInfo.nPage = rc.right-rc.left;

	// Scroll to 0 before we loose the scroll bars  
	if(g_lpInst->hsInfo.nMax < g_lpInst->hsInfo.nPage)
	{
		for(;0 < (g_lpInst->hsInfo.nPos--);Scroll(hwnd,1,0));
		g_lpInst->hsInfo.nPos = 0;
	}
	SetScrollInfo(hwnd,SB_HORZ,&g_lpInst->hsInfo,TRUE);

	g_lpInst->vsInfo.fMask = SIF_ALL;
	g_lpInst->vsInfo.cbSize = sizeof(SCROLLINFO);
	g_lpInst->vsInfo.nMin = 0;
	g_lpInst->vsInfo.nMax = prc->bottom;
	g_lpInst->vsInfo.nPage = rc.bottom-rc.top;

	if(g_lpInst->vsInfo.nMax < g_lpInst->vsInfo.nPage)
	{
		// Now make sure the header stays in view
		HWND hHeader = ListView_GetHeader(g_lpInst->hwndList);

		for(;0 < (g_lpInst->vsInfo.nPos--);Scroll(hwnd,0,1))
		{
			GetClientRect(hHeader,&rc);
			MoveWindow(hHeader,rc.left, g_lpInst->vsInfo.nPos,
				 rc.right - rc.left, rc.bottom-rc.top , TRUE);
		}
		g_lpInst->vsInfo.nPos = 0;
	}
	SetScrollInfo(hwnd,SB_VERT,&g_lpInst->vsInfo,TRUE);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnHScroll
*
*     PURPOSE:  Handle the horizontal scrollbar.
*
*     PARAMS:   HWND hwnd		- This window
*               HWND hwndCtl	- Handle of the scrollbar
*               UINT code		- users scrolling request
*               INT  pos		- scroll box (or thumb) position
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

void Grid_OnHScroll(HWND hwnd, HWND hwndCtl, UINT code, int pos)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);
	RECT rc;
	GetClientRect(hwnd,&rc);
	int line = ListView_GetFirstVisibleColWidth(g_lpInst->hwndList, &rc);
	int page = line * ((int)(rc.right-rc.left) / line);

	int oldPos = g_lpInst->hsInfo.nPos;

	switch (code)
	{
		case SB_BOTTOM:			//Scrolls to the lower right.
			g_lpInst->hsInfo.nPos = g_lpInst->hsInfo.nMax;
			break;

		case SB_TOP:			//Scrolls to the lower left.
			g_lpInst->hsInfo.nPos = g_lpInst->hsInfo.nMin;
			break;

		case SB_ENDSCROLL:		//Ends scroll.
			break;

		case SB_LINELEFT:		//Scrolls left by one unit.
			g_lpInst->hsInfo.nPos -= line;
			break;

		case SB_LINERIGHT:		//Scrolls right by one unit.
			g_lpInst->hsInfo.nPos += line;
			break;

		case SB_PAGELEFT:		//Scrolls left by the width of the window.
			g_lpInst->hsInfo.nPos -= page;
			break;

		case SB_PAGERIGHT:		//Scrolls right by the width of the window.
			g_lpInst->hsInfo.nPos += page;
			break;

		case SB_THUMBPOSITION:	//Scrolls to the absolute position. The current position is specified by the Pos parameter.
			g_lpInst->hsInfo.nPos = pos;
			break;

		case SB_THUMBTRACK:		//Drags scroll box to the specified position. The current position is specified by the Pos parameter.
			g_lpInst->hsInfo.nPos = pos;	
			break;				
	}
	// Ensure that we do not scroll beyond visible boundaries
	g_lpInst->hsInfo.nPos = g_lpInst->hsInfo.nPos < g_lpInst->hsInfo.nMin ||
		(code == SB_LINELEFT && g_lpInst->hsInfo.nPos < line) ?
			 g_lpInst->hsInfo.nMin : g_lpInst->hsInfo.nPos;
	g_lpInst->hsInfo.nPos = g_lpInst->hsInfo.nPos > g_lpInst->hsInfo.nMax - g_lpInst->hsInfo.nPage ?
		 g_lpInst->hsInfo.nMax - g_lpInst->hsInfo.nPage : g_lpInst->hsInfo.nPos;

	g_lpInst->hsInfo.fMask = SIF_POS;
	SetScrollInfo(hwnd,SB_HORZ,&g_lpInst->hsInfo,TRUE);
	Scroll(hwnd,-(g_lpInst->hsInfo.nPos-oldPos),0);
}

/****************************************************************************
*
*     FUNCTION: Grid_OnVScroll
*
*     PURPOSE:  Handle the vertical scrollbar.
*
*     PARAMS:   HWND hwnd		- This window
*               HWND hwndCtl	- Handle of the scrollbar
*               UINT code		- users scrolling request
*               INT  pos		- scroll box (or thumb) position
*
*     RETURNS:  VOID
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static VOID Grid_OnVScroll(HWND hwnd, HWND hwndCtl, UINT code, INT pos)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);
	RECT rc;
	GetClientRect(hwnd,&rc);
	int line = ListView_GetRowHeight(g_lpInst->hwndList, &rc);
	int page = line * ((int)(rc.bottom-rc.top) / line);

	pos = line * (pos / line);//ScrollPos in increments of line

	int oldPos = g_lpInst->vsInfo.nPos;

	switch (code)
	{
		case SB_BOTTOM:		//Scrolls to the lower right.
			g_lpInst->vsInfo.nPos = g_lpInst->vsInfo.nMax;
			break;

		case SB_TOP:		//Scrolls to the upper left.
			g_lpInst->vsInfo.nPos = g_lpInst->vsInfo.nMin;
			break;

		case SB_ENDSCROLL:	//Ends scroll.
			break;

		case SB_LINEDOWN:	//Scrolls one line down.
			g_lpInst->vsInfo.nPos += line;
			break;

		case SB_LINEUP:		//Scrolls one line up.
			g_lpInst->vsInfo.nPos -= line;
			break;

		case SB_PAGEDOWN:	//Scrolls one page down.
			g_lpInst->vsInfo.nPos += page;
			break;

		case SB_PAGEUP:		//Scrolls one page up.
			g_lpInst->vsInfo.nPos -= page;
			break;

		case SB_THUMBPOSITION:	//Scrolls to the absolute position. The current position is specified by the nPos parameter.
			g_lpInst->vsInfo.nPos = pos;
			break;

		case SB_THUMBTRACK:	//Drags scroll box to the specified position. The current position is specified by the nPos parameter.
			g_lpInst->vsInfo.nPos = pos;
			break;							
	}
	g_lpInst->vsInfo.nPos = g_lpInst->vsInfo.nPos < g_lpInst->vsInfo.nMin ? g_lpInst->vsInfo.nMin : g_lpInst->vsInfo.nPos;
	g_lpInst->vsInfo.nPos = g_lpInst->vsInfo.nPos > g_lpInst->vsInfo.nMax - g_lpInst->vsInfo.nPage ? g_lpInst->vsInfo.nMax - g_lpInst->vsInfo.nPage : g_lpInst->vsInfo.nPos;

	// In order to reduce flicker in the header we position the header
    // before or after the scroll depending on the direction of scroll

	if(oldPos < pos) PinHeaderToTop(ListView_GetHeader(g_lpInst->hwndList)); 

	g_lpInst->vsInfo.fMask = SIF_POS;
	SetScrollInfo(hwnd,SB_VERT,&g_lpInst->vsInfo,TRUE);
	Scroll(hwnd,0,-(g_lpInst->vsInfo.nPos-oldPos));

	if(oldPos > pos)PinHeaderToTop(ListView_GetHeader(g_lpInst->hwndList));
}

/****************************************************************************
*
*     FUNCTION: Grid_OnSetCursor
*
*     PURPOSE:  The cursor changes when the mouse leaves the header, this is the
*                only easy indication that a column resize has been compleated.
*                I use the event to crop the grid back to its data and to reposition
*                the editor.
*
*     PARAMS:   HWND hwnd - handle of control
*               HWND hwndCursor - the handle of the cursor
*               UINT codeHitTest - hit test code
*               UINT msg - error message related to hit test
*
*     RETURNS:  BOOL - TRUE if handled
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

BOOL Grid_OnSetCursor(HWND hwnd, HWND hwndCursor, UINT codeHitTest, UINT msg)
{
	//
	// In Grid_OnColumnResize() we set g_fsizeCol when we clicked on the listview header;
	//  now we have a means of knowing that the uncaptured cursor has
	//  just slipped off the header. 
	//
	if(g_fsizeCol)
	{
		RECT rc;
		Grid_GetInstanceData(hwnd,&g_lpInst);

		Grid_OnSize(hwnd, 0, 0, 0);

		//RePosition the editor
		ListView_GetSubItemRect(g_lpInst->hwndList,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,LVIR_LABEL,&rc);
		MapWindowPoints(g_lpInst->hwndList, hwnd, (LPPOINT)&rc.left,2);
		MoveWindow(g_lpInst->hwndEditor,
			rc.left,rc.top,rc.right-rc.left,rc.bottom-rc.top,TRUE);

		//Show the editor
		ShowWindow(g_lpInst->hwndEditor,TRUE);

		g_fsizeCol = FALSE;
	}
	return FALSE; //Since this is not the standard use of this event don't handle it.
}

/****************************************************************************
*
*     FUNCTION: Grid_OnColumnResize
*
*     PURPOSE:  Handles the HDN_ITEMCHANGING notification due to a resize attempt.
*
*     PARAMS:   HWND hwnd - handle of control
*               LPNMHDR pnm - pointer to the notification message header
*
*     RETURNS:  BOOL - TRUE = non resizeable, FALSE = do resize
*
* History:
*               Sep '07 - Created
*
\****************************************************************************/

BOOL Grid_OnColumnResize(HWND hwnd, LPNMHDR pnm)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);

	if(!g_lpInst->fResizableHeader) return TRUE;

	ShowWindow(g_lpInst->hwndEditor,FALSE);
	g_fsizeCol = TRUE;

	ListView_Resize(g_lpInst->hwndList);

	return FALSE;
}

/****************************************************************************
*
*     FUNCTION: Grid_OnCustomDraw
*
*     PURPOSE:  Define the look and feel of the Grid.
*
*     PARAMS:   HWND hwnd - handle of control
*               LPNMLVCUSTOMDRAW lplvcd - pointer to a custom draw message struct
*
*     RETURNS:  LRESULT - depends on custom draw notification
*
* History:
*                Sep '07 - Created
*
\****************************************************************************/

LRESULT Grid_OnCustomDraw (HWND hwnd, LPNMLVCUSTOMDRAW lplvcd)
{
	Grid_GetInstanceData(hwnd,&g_lpInst);

    switch(lplvcd->nmcd.dwDrawStage) 
    {
		//Before the paint cycle begins
		//request notifications for individual listview items
        case CDDS_PREPAINT : return CDRF_NOTIFYITEMDRAW;
 
		//Before an item is drawn
 		//request notifications for individual listview subitems
        case CDDS_ITEMPREPAINT: return CDRF_NOTIFYSUBITEMDRAW;

        case CDDS_SUBITEM | CDDS_ITEMPREPAINT:
        {
			// Color alternating Rows
			if(g_lpInst->fPaintByRow&&lplvcd->nmcd.dwItemSpec & 1)
			{
				lplvcd->clrText = g_lpInst->Alt_TxtColr;
				lplvcd->clrTextBk = g_lpInst->Alt_BkColr;
			}
			// Color alternating Cols
			else if(!g_lpInst->fPaintByRow&&lplvcd->iSubItem & 1)
			{
            	lplvcd->clrText   = g_lpInst->Alt_TxtColr;
            	lplvcd->clrTextBk = g_lpInst->Alt_BkColr;
			}
			else // Color default
			{
            	lplvcd->clrText   = ListView_GetTextColor(g_lpInst->hwndList);
            	lplvcd->clrTextBk = ListView_GetBkColor(g_lpInst->hwndList);
			}
			// Color active cell
			if( lplvcd->iSubItem == g_lpInst->hti.iSubItem &&
				lplvcd->nmcd.dwItemSpec == g_lpInst->hti.iItem)
			{
            	lplvcd->clrText   = g_lpInst->Editor_TxtColr;
            	lplvcd->clrTextBk = g_lpInst->Editor_BkColr;
			}
			// Skin Column headers
			if(g_lpInst->fRowHeaders)
			{
				HWND hHdr;
				HDITEM hdi;
				RECT rc;
				HDC hdc;

				hHdr = ListView_GetHeader(hwnd);
				Header_GetItem(hHdr,lplvcd->iSubItem, &hdi);
				Header_GetItemRect(hHdr,lplvcd->iSubItem, &rc);
				hdc=GetDC(hHdr);

				if(lplvcd->iSubItem == g_lpInst->hti.iSubItem)
					DrawEdge(hdc, &rc,EDGE_SUNKEN,BF_RECT|BF_SOFT|BF_MIDDLE);
				else
					DrawEdge(hdc, &rc,EDGE_RAISED,BF_RECT|BF_SOFT|BF_MIDDLE);

				rc.left+=6;
				rc.bottom-=2;

				SetBkMode(hdc,TRANSPARENT);
				SetTextColor(hdc,GetSysColor(COLOR_BTNTEXT));
				SelectObject(hdc,GetStockObject(DEFAULT_GUI_FONT));
				DrawText(hdc,hdi.pszText,lstrlen(hdi.pszText), &rc,DT_SINGLELINE|DT_BOTTOM|DT_LEFT);
				SetBkMode(hdc,OPAQUE);

				ReleaseDC(g_lpInst->hwndList,hdc);
			}
			// Draw Row headers
			if(lplvcd->iSubItem == 0&&g_lpInst->fRowHeaders)
			{
				static char buf [2048];
				RECT  rc;
				HDC  hdc;

				ListView_GetSubItemRect(g_lpInst->hwndList,
					lplvcd->nmcd.dwItemSpec,0,LVIR_LABEL,&rc);

				rc.left-=2;
				rc.top-=1;
				rc.bottom-=1;

				hdc=GetDC(g_lpInst->hwndList);

				if(lplvcd->nmcd.dwItemSpec==g_lpInst->hti.iItem)
					DrawEdge(hdc, &rc,EDGE_SUNKEN,BF_RECT|BF_SOFT|BF_MIDDLE);
				else
					DrawEdge(hdc, &rc,EDGE_RAISED,BF_RECT|BF_SOFT|BF_MIDDLE);

				ListView_GetItemText(g_lpInst->hwndList,lplvcd->nmcd.dwItemSpec,0,buf,sizeof buf);

				SetBkMode(hdc,TRANSPARENT);
				SetTextColor(hdc,GetSysColor(COLOR_BTNTEXT));
				SelectObject(hdc,GetStockObject(DEFAULT_GUI_FONT));
				DrawText(hdc,buf,lstrlen(buf), &rc,DT_SINGLELINE|DT_VCENTER|DT_CENTER);
				SetBkMode(hdc,OPAQUE);

				ReleaseDC(g_lpInst->hwndList,hdc);

				return CDRF_SKIPDEFAULT;
			}
			return CDRF_NEWFONT;
		}
    }
    return CDRF_DODEFAULT;
}

/****************************************************************************
*
*     FUNCTION: Grid_OnMouseClick
*
*     PURPOSE:  Handles mouse clicks on the ListView.
*
*     PARAMS:   HWND hwnd - handle of control
*               LPNMHDR pnm - pointer to the notification message header
*
*     RETURNS:  BOOL - TRUE if handled
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

BOOL Grid_OnMouseClick(HWND hwnd, LPNMHDR pnm)
{
	static char buf[2048];
	Grid_GetInstanceData(hwnd,&g_lpInst);

	//Update the grid's currently selected subItem before moving on
	Edit_GetText(g_lpInst->hwndEditor, buf, sizeof buf);
	ListView_SetItemText(pnm->hwndFrom,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,buf);
	ShowWindow(g_lpInst->hwndEditor,FALSE);
	
	//Get the mouse position at time NM_CLICK sent
	DWORD dwpos = GetMessagePos();
	g_lpInst->hti.pt.x = GET_X_LPARAM(dwpos);
	g_lpInst->hti.pt.y = GET_Y_LPARAM(dwpos);
	MapWindowPoints(HWND_DESKTOP,pnm->hwndFrom, &g_lpInst->hti.pt,1);

	//Get the Item, SubItem, and SubItem rectangle
	RECT rc;
	INT iSubItemOld = g_lpInst->hti.iSubItem > 0 ? g_lpInst->hti.iSubItem : 1;
	INT iItemOld = g_lpInst->hti.iItem < 0 ? 0 : g_lpInst->hti.iItem;

	ListView_SubItemHitTest(pnm->hwndFrom, &g_lpInst->hti);
	if(g_lpInst->hti.iItem==-1) return FALSE;

	if(LVN_COLUMNCLICK==pnm->code) //Adjust ItemOld if it is off screen
	{
		GetClientRect(hwnd,&rc);

		if(g_lpInst->hti.iItem > iItemOld ||
		   ListView_GetLastVisibleRow(g_lpInst->hwndList,&rc) < iItemOld)
			iItemOld = g_lpInst->hti.iItem +1;
	}
	
	//Click a row header Keep old sub item
	if(g_lpInst->fRowHeaders && g_lpInst->hti.iSubItem == 0 &&
		LVN_COLUMNCLICK!=pnm->code)
	{
		g_lpInst->hti.iSubItem = iSubItemOld;

		// Update ListView if selected sub item out of the window
		ListView_RedrawItems(hwnd,
			(iItemOld < g_lpInst->hti.iItem ? iItemOld : g_lpInst->hti.iItem),
			(iItemOld > g_lpInst->hti.iItem ? iItemOld : g_lpInst->hti.iItem));
		UpdateWindow(hwnd);
    }

	//Click a column header Keep old item
	if(LVN_COLUMNCLICK==pnm->code)
	{
		if(0 == g_lpInst->hti.iSubItem && g_lpInst->fRowHeaders)
		{
			//If using Row headers then clicking col 0 places the edit box
			// over the upper left cell 
			g_lpInst->hti.iSubItem = 1;
			g_lpInst->hti.iItem += g_lpInst->hti.iItem > 0 ? 1 : 0;
		}
		else g_lpInst->hti.iItem = iItemOld;
	}

	ListView_GetSubItemRect(pnm->hwndFrom,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,LVIR_LABEL,&rc);

	//Map the SubItem rectangle to the parent dialog so that
	//we can accurately position the Edit.
	MapWindowPoints(pnm->hwndFrom, hwnd, (LPPOINT)&rc.left,2);
	SetWindowPos(g_lpInst->hwndEditor,
		NULL,rc.left,rc.top, rc.right-rc.left,rc.bottom-rc.top,SWP_NOZORDER|SWP_DEFERERASE);

	//Get the SubItem Text
	ListView_GetItemText(pnm->hwndFrom,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,buf,sizeof buf);

	//Load and display editor
	Edit_SetText(g_lpInst->hwndEditor, buf);
	Edit_SetSel(g_lpInst->hwndEditor,0,strlen(buf));
	ShowWindow(g_lpInst->hwndEditor,TRUE);

	return TRUE;
}

/****************************************************************************
*
*     FUNCTION: Grid_OnNotify
*
*     PURPOSE:  Handles notification messages for control's components.
*
*     PARAMS:   HWND hwnd - handle of control
*               INT id - id of component
*               LPNMHDR pnm - pointer to the notification message header
*
*     RETURNS:  BOOL - TRUE if handled
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static BOOL  Grid_OnNotify(HWND hwnd, INT id, LPNMHDR pnm)
{
	//
	// If this is a resizable grid, initiate column resizing
	//  resizing is concluded in the Grid_OnSetCursor event handler
	//
	if(HDN_ITEMCHANGING==pnm->code)
		return Grid_OnColumnResize(hwnd, pnm);
	//
	// Process Custom draw
	//
	else if(ID_LISTVIEW==id&&pnm->code == NM_CUSTOMDRAW)
		return Grid_OnCustomDraw(hwnd, (LPNMLVCUSTOMDRAW)pnm);
	//
	// Handle mouse clicks on the grid
	//
	else if(ID_LISTVIEW==id)
	{
		Grid_GetInstanceData(hwnd,&g_lpInst);
		if((NM_DBLCLK==pnm->code&&g_lpInst->fDblClick)||
			(NM_CLICK==pnm->code&&!g_lpInst->fDblClick)||
			(LVN_COLUMNCLICK==pnm->code))
			return Grid_OnMouseClick(hwnd, pnm);
		else
			return FALSE;
	}
	else
		return FALSE;
}

/****************************************************************************
*
*     FUNCTION: ListView_KeyProc
*
*     PURPOSE:  Window Procedure for the ListView.
*
*    COMMENTS:  Handles Keyboard input to the listView; mostly
*                moving the editor arround.
*
*      PARAMS:  HWND   hList	- Handle of ListView
*               UINT   msg		- Which message?
*               WPARAM wParam	- message parameter
*               LPARAM lParam	- message parameter
*
*     RETURNS:  LRESULT CALLBACK - depends on message
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static LRESULT CALLBACK ListView_KeyProc (HWND hList, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Note: Instance data is attached to ListView's parent
	static RECT rc, rcParent;
	static char buf[2048];

	if(WM_DESTROY==msg) // UnSubclass the ListView Control
	{
		SetWindowLong(hList,GWL_WNDPROC,(DWORD)GetProp(hList,"Wprc"));
		RemoveProp(hList,"Wprc");
		return 0;
	}
	else if(WM_KEYDOWN==msg)
	{
		HWND hParent = GetParent(hList);

		// If the column is in resize mode when a key is pressed
		//  call the code to end column resizing first
		if(g_fsizeCol) Grid_OnSetCursor(hParent,NULL,0,0);

		Grid_GetInstanceData(hParent, &g_lpInst);
		GetClientRect(hParent, &rcParent);

		// iItem and iSubitem might still be initialized to -1
		//  if this is the case, select the first cell.
		if(g_lpInst->hti.iItem == -1) g_lpInst->hti.iItem = 0;
		if(g_lpInst->hti.iSubItem == -1) g_lpInst->hti.iSubItem = 0;

		switch(wParam)
		{
			case VK_HOME:
				if(g_lpInst->fRowHeaders) g_lpInst->hti.iSubItem = 1;
				else g_lpInst->hti.iSubItem = 0;
				break;
			case VK_END:
				g_lpInst->hti.iSubItem = Grid_GetColumnCount(hList);
				break;
			case VK_PRIOR:
				g_lpInst->hti.iItem = 0;
				break;
			case VK_NEXT:
				g_lpInst->hti.iItem = ListView_GetItemCount(hList) -1;
				break;
			case VK_UP:
				g_lpInst->hti.iItem = 0 == g_lpInst->hti.iItem ? 0 : g_lpInst->hti.iItem -1;
 				break;
			case VK_DOWN:
				g_lpInst->hti.iItem = (ListView_GetItemCount(hList)-1) == g_lpInst->hti.iItem ?
					ListView_GetItemCount(hList) -1 : g_lpInst->hti.iItem +1;
				break;
			case VK_LEFT:
				if(g_lpInst->fRowHeaders)
					g_lpInst->hti.iSubItem = 1 >= g_lpInst->hti.iSubItem ?
						1 : g_lpInst->hti.iSubItem-1;
				else
					g_lpInst->hti.iSubItem = 0 == g_lpInst->hti.iSubItem ?
						0 : g_lpInst->hti.iSubItem-1;
				break;
			case VK_RIGHT:
				g_lpInst->hti.iSubItem = Grid_GetColumnCount(hList) == g_lpInst->hti.iSubItem ?
					Grid_GetColumnCount(hList) : g_lpInst->hti.iSubItem+1;
				break;
			default:
				return 0;
		}
		//Position the editor
		ShowWindow(g_lpInst->hwndEditor,FALSE);
		ListView_GetSubItemRect(hList,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,LVIR_LABEL,&rc);
		MapWindowPoints(hList, hParent, (LPPOINT)&rc.left,2);
		SetWindowPos(g_lpInst->hwndEditor,NULL,rc.left,rc.top,
			rc.right-rc.left,rc.bottom-rc.top,SWP_NOZORDER|SWP_DEFERERASE);

		//Get the SubItem Text
		ListView_GetItemText(hList,g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,buf,sizeof buf);

		//Load and display editor
		Edit_SetText(g_lpInst->hwndEditor, buf);
		ShowWindow(g_lpInst->hwndEditor,TRUE);

		//Trigger an NM_CUSTOMDRAW notification for the curent liststview item
		// in order to update the row header to show current row selection

		ListView_RedrawItems(hList,g_lpInst->hti.iItem-1,g_lpInst->hti.iItem+1);
		UpdateWindow(g_lpInst->hwndList);

	    //Adjust Scrolls
		switch(wParam)
		{
			case VK_HOME:
					SNDMSG(hParent,WM_HSCROLL,SB_TOP,0);
				break;
			case VK_END:
				if(!PtInRect(&rcParent,*((LPPOINT)&rc.right)))
					SNDMSG(hParent,WM_HSCROLL,SB_BOTTOM,0);
				break;
			case VK_PRIOR:
					SNDMSG(hParent,WM_VSCROLL,SB_TOP,0);
				break;
			case VK_NEXT:
				if(!PtInRect(&rcParent,*((LPPOINT)&rc.right)))
					SNDMSG(hParent,WM_VSCROLL,SB_BOTTOM,0);
				break;
			case VK_UP:
			{
				RECT rcHeader;
				GetClientRect(ListView_GetHeader(hList),&rcHeader);

				//Keep the edit box from overlaying the header bar
				rcParent.top += (rcHeader.bottom-rcHeader.top);

				if(!PtInRect(&rcParent,*((LPPOINT)&rc.left)))
					SNDMSG(hParent,WM_VSCROLL,SB_LINEUP,0);
			}
				break;
			case VK_DOWN:
				rcParent.bottom -= ListView_GetRowHeight(hList, &rcParent);
				if(!PtInRect(&rcParent,*((LPPOINT)&rc.left)))
					SNDMSG(hParent,WM_VSCROLL,SB_LINEDOWN,0);
				break;
			case VK_LEFT:
				rcParent.left += ListView_GetFirstVisibleColWidth(hList, &rcParent);
				if(!PtInRect(&rcParent,*((LPPOINT)&rc.right)))
					SNDMSG(hParent,WM_HSCROLL,SB_LINELEFT,0);
				break;
			case VK_RIGHT:
				rcParent.right -= ListView_GetLastVisibleColWidth(hList, &rcParent);
				if(!PtInRect(&rcParent,*((LPPOINT)&rc.left)))
					SNDMSG(hParent,WM_HSCROLL,SB_LINERIGHT,0);
				break;
		}
		return 0;
	}
	else if(WM_KEYUP==msg && VK_RETURN==wParam)// Edit subItem
	{
		SNDMSG(g_lpInst->hwndEditor,WM_KEYDOWN,(WPARAM)VK_END,0L);
		//SNDMSG(g_lpInst->hwndEditor,WM_CHAR,wParam,0L);
		ShowWindow(g_lpInst->hwndEditor,TRUE);
		SetFocus(g_lpInst->hwndEditor);
	}
	return CallWindowProc((WNDPROC)GetProp(hList,(LPCSTR)"Wprc"),hList,msg,wParam,lParam);
}

/****************************************************************************
*
*     FUNCTION: Editor_KeyProc
*
*     PURPOSE:  Window Procedure for the Editor.
*
*    COMMENTS:  SubClassing gives access to Editor's WM_CHAR and WM_KEYDOWN
*                messages.  I handle the Enter (Return) key as WM_CHAR in
*                order to disable the bell.  I handle the Esc key in WM_KEYDOWN
*                so that the default behavior (initiating WM_DESTROY) is overridden.
*                This behavior seems to take place between WM_KEYDOWN and WM_CHAR
*                stages of message processing.
*
*      PARAMS:  HWND   hEdit	- Handle of Editor
*               UINT   msg		- Which message?
*               WPARAM wParam	- message parameter
*               LPARAM lParam	- message parameter
*
*     RETURNS:  LRESULT CALLBACK - depends on message
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static LRESULT CALLBACK Editor_KeyProc (HWND hEdit, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static char buf[2048];

	if(WM_DESTROY==msg) // UnSubclass the Edit Control
	{
		SetWindowLong(hEdit,GWL_WNDPROC,(DWORD)GetProp(hEdit,"Wprc"));
		RemoveProp(hEdit,"Wprc");
		return 0;
	}
	else if(WM_CHAR==msg&&VK_RETURN==wParam)  
		return TRUE; // handle Enter (NO BELL) 

	else if(WM_KEYUP==msg && (VK_RETURN==wParam || VK_RIGHT==wParam ||
		 VK_LEFT==wParam || VK_UP==wParam || VK_DOWN==wParam))
	{
		// Note: Instance data is attached to Edit's parent
		Grid_GetInstanceData(GetParent(hEdit),&g_lpInst);
		char temp [Edit_GetTextLength(hEdit)+1];
		Edit_GetText(hEdit, temp, sizeof temp);
		ListView_SetItemText(g_lpInst->hwndList,
			g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,temp);
		SetFocus(g_lpInst->hwndList);
		Refresh(hEdit);

		if(VK_RETURN!=wParam)//Go to next cell in edit mode
		{
			SNDMSG(g_lpInst->hwndList,WM_KEYDOWN,(WPARAM)wParam,0L);
			SNDMSG(g_lpInst->hwndList,WM_KEYUP,(WPARAM)VK_RETURN,0L);
		}
		return 0;
	}
	else if(WM_KEYDOWN==msg&&VK_ESCAPE==wParam) // handle Escape (NO WM_DESTROY)
	{
		Grid_GetInstanceData(GetParent(hEdit),&g_lpInst);
		//Get the SubItem Text
		ListView_GetItemText(g_lpInst->hwndList,
			g_lpInst->hti.iItem,g_lpInst->hti.iSubItem,buf,sizeof buf);
		//Reset the editor
		Edit_SetText(hEdit, buf);
		//Focus back to listview
		SetFocus(g_lpInst->hwndList);
		return 0;
	}
 	return CallWindowProc((WNDPROC)GetProp(hEdit,(LPCSTR)"Wprc"),hEdit,msg,wParam,lParam);
}

/****************************************************************************
*
*     FUNCTION: ListViewHeader_MoveProc
*
*     PURPOSE:  Window Procedure for the ListView Header.
*
*    COMMENTS:  Override the WM_MOVE event of the header.
*
*      PARAMS:  HWND   hHeader	- Handle of ListView's Header
*               UINT   msg		- Which message?
*               WPARAM wParam	- message parameter
*               LPARAM lParam	- message parameter
*
*     RETURNS:  LRESULT CALLBACK - depends on message
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static LRESULT CALLBACK ListViewHeader_MoveProc (HWND hHeader, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if(WM_DESTROY==msg) // UnSubclass the ListView Header
	{
		SetWindowLong(hHeader,GWL_WNDPROC,(DWORD)GetProp(hHeader,"Wprc"));
		RemoveProp(hHeader,"Wprc");
		return 0;
	}
	else if (WM_MOVE==msg)
	{
		// Note: Instance data is attached to Header's grandparent
		Grid_GetInstanceData(GetParent(GetParent(hHeader)),&g_lpInst);

		// Now make sure the header stays in view
		SetWindowPos(hHeader,NULL,0,g_lpInst->vsInfo.nPos,
			0,0,SWP_NOZORDER|SWP_NOSIZE|SWP_DEFERERASE);
	}
	return CallWindowProc((WNDPROC)GetProp(hHeader,(LPCSTR)"Wprc"),hHeader,msg,wParam,lParam);
}

/****************************************************************************
*
*     FUNCTION: Grid_WndProc
*
*     PURPOSE:  Window Procedure for the Data View Grid.
*
*     PARAMS:   HWND   hwnd		- This window
*               UINT   msg		- Which message?
*               WPARAM wParam	- message parameter
*               LPARAM lParam	- message parameter
*
*     RETURNS:  LRESULT CALLBACK - depends on message
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

static LRESULT CALLBACK Grid_WndProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Pass along any list view specific messages transparently
	if(LVM_FIRST <= msg && msg <= LVM_FIRST + 181)
	{
		Grid_GetInstanceData(hwnd,&g_lpInst);
		return SNDMSG(g_lpInst->hwndList,msg,wParam,lParam);
	}
	switch(msg)
	{
		HANDLE_MSG (hwnd, WM_CTLCOLOREDIT, Grid_OnCtlColorEdit);
		HANDLE_MSG (hwnd, WM_CREATE, Grid_OnCreate);
		HANDLE_MSG (hwnd, WM_DESTROY, Grid_OnDestroy);
		HANDLE_MSG (hwnd, WM_GETDLGCODE, Grid_OnGetDlgCode);
		HANDLE_MSG (hwnd, WM_SIZE, Grid_OnSize);
		HANDLE_MSG (hwnd, WM_HSCROLL, Grid_OnHScroll);
		HANDLE_MSG (hwnd, WM_VSCROLL, Grid_OnVScroll);
		HANDLE_MSG (hwnd, WM_NOTIFY, Grid_OnNotify);
		HANDLE_MSG (hwnd, WM_SETCURSOR, Grid_OnSetCursor);

		case DGVM_GETLISTVIEWCONTROL:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->hwndList;

		case DGVM_GETEDITCONTROL:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->hwndEditor;

		case DGVM_GETCOLUMNCOUNT:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) Grid_GetColumnCount(g_lpInst->hwndList);

		case DGVM_GETEDITORBACKCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->Editor_BkColr;

		case DGVM_SETEDITORBACKCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->Editor_BkColr = (COLORREF)lParam;
			return TRUE;

		case DGVM_GETEDITORTEXTCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->Editor_TxtColr;

		case DGVM_SETEDITORTEXTCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->Editor_TxtColr = (COLORREF)lParam;
			return TRUE;

		case DGVM_RESIZEABLEHEADER:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->fResizableHeader = (BOOL)wParam;
			return TRUE;

		case DGVM_GETALTBACKCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->Alt_BkColr;

		case DGVM_SETALTBACKCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->Alt_BkColr = (COLORREF)lParam;
			g_lpInst->fPaintByRow = (BOOL) wParam;
			return TRUE;

		case DGVM_SETALTTXTCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->Alt_TxtColr = (COLORREF)lParam;
			return TRUE;

		case DGVM_GETALTTXTCLR:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			return (LRESULT) g_lpInst->Alt_TxtColr;

		case DGVM_SETDOUBLECLICKSELECT:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->fDblClick = (BOOL) wParam;
			return TRUE;

		case DGVM_FIRSTCOLUMNROWHEADERS:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			g_lpInst->fRowHeaders = (BOOL) wParam;
			Refresh(g_lpInst->hwndList);
			return TRUE;

		case DGVM_SETROWHEIGHT:
			Grid_GetInstanceData(hwnd,&g_lpInst);
			Grid_SetRowHeight(g_lpInst->hwndList, (INT) wParam);
			return TRUE;

		default: return DefWindowProc (hwnd, msg, wParam, lParam);
	}
	//return 0;
}

/****************************************************************************
*
*     FUNCTION: InitDataGridView
*
*     PURPOSE:  Initialize and register the custom control
*
*     PARAMS:   HINSTANCE hInstance - handle of application instance
*
*     RETURNS:  BOOL TRUE if successfull
*
* History:
*                Aug '07 - Created
*
\****************************************************************************/

BOOL InitDataGridView(HINSTANCE hInstance)
{
	//TODO: Create a Grid handle struct for multiple instances

    WNDCLASSEX wcex;
    ATOM aReturn;

    wcex.cbSize          = sizeof(WNDCLASSEX);
    wcex.style           = CS_BYTEALIGNCLIENT;
    wcex.lpfnWndProc     = (WNDPROC)Grid_WndProc;
    wcex.cbClsExtra      = 0;
    wcex.cbWndExtra      = 0;
    wcex.hInstance       = hInstance;
    wcex.hCursor         = NULL;
    wcex.hbrBackground   = (HBRUSH)(GetStockObject(GRAY_BRUSH));//(COLOR_WINDOW + 1);
    wcex.lpszMenuName    = NULL;
    wcex.lpszClassName   = g_szClassName;
    wcex.hIcon           = NULL;
    wcex.hIconSm         = NULL;

	aReturn = RegisterClassEx(&wcex);

    if (0 == aReturn)
    {
        WNDCLASS wc;

        wc.style          = CS_BYTEALIGNCLIENT;
        wc.lpfnWndProc    = (WNDPROC)Grid_WndProc;
        wc.cbClsExtra     = 0;
        wc.cbWndExtra     = 0;
        wc.hInstance      = hInstance;
        wc.hIcon          = NULL;
        wc.hCursor        = NULL;
        wc.hbrBackground  = (HBRUSH)(GetStockObject(GRAY_BRUSH));//(COLOR_WINDOW + 1);
        wc.lpszMenuName   = NULL;
        wc.lpszClassName  = g_szClassName;

        aReturn = RegisterClass(&wc);
    }

    return aReturn;
}
